"""
This type stub file was generated by pyright.
"""

import re
import yaml
import toml
from types import ModuleType
from typing import Any, TYPE_CHECKING, Type
from frontmatter import Post

"""
.. testsetup:: handlers

    import frontmatter

By default, ``frontmatter`` reads and writes YAML metadata. But maybe
you don't like YAML. Maybe enjoy writing metadata in JSON, or TOML, or
some other exotic markup not yet invented. For this, there are handlers.

This module includes handlers for YAML, JSON and TOML, as well as a 
:py:class:`BaseHandler <frontmatter.default_handlers.BaseHandler>` that 
outlines the basic API and can be subclassed to deal with new formats.

**Note**: The TOML handler is only available if the `toml <https://pypi.org/project/toml/>`_
library is installed.

Handlers
--------

Handlers do most of the underlying work parsing and exporting front matter.
When you call :py:func:`frontmatter.loads <frontmatter.loads>`, frontmatter first needs to figure out the
best handler for the format you're using (YAML, JSON, TOML, etc), then call
methods to read or write metadata.

A handler needs to do four things:

- detect whether it can parse the given piece of text
- split front matter from content, returning both as a two-tuple
- parse front matter into a Python dictionary
- export a dictionary back into text

An example:

Calling :py:func:`frontmatter.load <frontmatter.load>` (or :py:func:`loads <frontmatter.loads>`) 
with the ``handler`` argument tells frontmatter which handler to use. 
The handler instance gets saved as an attribute on the returned post 
object. By default, calling :py:func:`frontmatter.dumps <frontmatter.dumps>` 
on the post will use the attached handler.


::

    >>> import frontmatter
    >>> from frontmatter.default_handlers import YAMLHandler, TOMLHandler
    >>> post = frontmatter.load('tests/toml/hello-toml.md', handler=TOMLHandler())
    >>> post.handler #doctest: +ELLIPSIS
    <frontmatter.default_handlers.TOMLHandler object at 0x...>

    >>> print(frontmatter.dumps(post)) # doctest: +SKIP
    +++
    test = "tester"
    something = "else"
    author = "bob"
    +++
    <BLANKLINE>
    Title
    =====
    <BLANKLINE>
    title2
    ------
    <BLANKLINE>
    Hello.
    <BLANKLINE>
    Just need three dashes
    ---
    <BLANKLINE>
    And this shouldn't break.

Passing a new handler to :py:func:`frontmatter.dumps <frontmatter.dumps>` 
(or :py:func:`dump <frontmatter.dump>`) changes the export format:

::

    >>> print(frontmatter.dumps(post, handler=YAMLHandler())) # doctest: +NORMALIZE_WHITESPACE
    ---
    author: bob
    something: else
    test: tester
    ---
    <BLANKLINE>
    Title
    =====
    <BLANKLINE>
    title2
    ------
    <BLANKLINE>
    Hello.
    <BLANKLINE>
    Just need three dashes
    ---
    <BLANKLINE>
    And this shouldn't break.

Changing the attached ``handler`` on a post has the same effect. Setting ``handler``
to ``None`` will default the post back to :py:class:`YAMLHandler <frontmatter.default_handlers.YAMLHandler>`.
These three variations will produce the same export:

::

    # set YAML format when dumping, but the old handler attached
    >>> t1 = frontmatter.dumps(post, handler=YAMLHandler())
    >>> post.handler = YAMLHandler() # set a new handler, changing all future exports
    >>> t2 = frontmatter.dumps(post)
    >>> post.handler = None # remove handler, defaulting back to YAML
    >>> t3 = frontmatter.dumps(post)
    >>> t1 == t2 == t3
    True

All handlers use the interface defined on ``BaseHandler``. Each handler needs to know how to:

- split metadata and content, based on a boundary pattern (``handler.split``)
- parse plain text metadata into a Python dictionary (``handler.load``)
- export a dictionary back into plain text (``handler.export``)
- format exported metadata and content into a single string (``handler.format``)


"""
SafeDumper: Type[yaml.CDumper] | Type[yaml.SafeDumper]
SafeLoader: Type[yaml.CSafeLoader] | Type[yaml.SafeLoader]
toml: ModuleType | None
if TYPE_CHECKING: ...
__all__ = ["BaseHandler", "YAMLHandler", "JSONHandler"]
if toml: ...
DEFAULT_POST_TEMPLATE = ...

class BaseHandler:
    """
    BaseHandler lays out all the steps to detecting, splitting, parsing and
    exporting front matter metadata.

    All default handlers are subclassed from BaseHandler.
    """

    FM_BOUNDARY: re.Pattern[str] | None = ...
    START_DELIMITER: str | None = ...
    END_DELIMITER: str | None = ...
    def __init__(
        self,
        fm_boundary: re.Pattern[str] | None = ...,
        start_delimiter: str | None = ...,
        end_delimiter: str | None = ...,
    ) -> None: ...
    def detect(self, text: str) -> bool:
        """
        Decide whether this handler can parse the given ``text``,
        and return True or False.

        Note that this is *not* called when passing a handler instance to
        :py:func:`frontmatter.load <frontmatter.load>` or :py:func:`loads <frontmatter.loads>`.
        """
        ...

    def split(self, text: str) -> tuple[str, str]:
        """
        Split text into frontmatter and content
        """
        ...

    def load(self, fm: str) -> dict[str, Any]:
        """
        Parse frontmatter and return a dict
        """
        ...

    def export(self, metadata: dict[str, object], **kwargs: object) -> str:
        """
        Turn metadata back into text
        """
        ...

    def format(self, post: Post, **kwargs: object) -> str:
        """
        Turn a post into a string, used in ``frontmatter.dumps``
        """
        ...

class YAMLHandler(BaseHandler):
    """
    Load and export YAML metadata. By default, this handler uses YAML's
    "safe" mode, though it's possible to override that.
    """

    FM_BOUNDARY = ...
    END_DELIMITER = ...
    def load(self, fm: str, **kwargs: object) -> Any:
        """
        Parse YAML front matter. This uses yaml.SafeLoader by default.
        """
        ...

    def export(self, metadata: dict[str, object], **kwargs: object) -> str:
        """
        Export metadata as YAML. This uses yaml.SafeDumper by default.
        """
        ...

class JSONHandler(BaseHandler):
    """
    Load and export JSON metadata.

    Note that changing ``START_DELIMITER`` or ``END_DELIMITER`` may break JSON parsing.
    """

    FM_BOUNDARY = ...
    START_DELIMITER = ...
    END_DELIMITER = ...
    def split(self, text: str) -> tuple[str, str]: ...
    def load(self, fm: str, **kwargs: object) -> Any: ...
    def export(self, metadata: dict[str, object], **kwargs: object) -> str:
        "Turn metadata into JSON"
        ...

if toml:
    class TOMLHandler(BaseHandler):
        """
        Load and export TOML metadata.

        By default, split based on ``+++``.
        """

        FM_BOUNDARY = ...
        END_DELIMITER = ...
        def load(self, fm: str, **kwargs: object) -> Any: ...
        def export(self, metadata: dict[str, object], **kwargs: object) -> str:
            "Turn metadata into TOML"
            ...

else:
    TOMLHandler: Type[TOMLHandler] | None = ...



from pathlib import Path
from typing import TypedDict, Any, cast
from typing_extensions import Annotated
from types import CodeType
from datetime import datetime
import ast
import typer
from rich.console import Console
import frontmatter
import jsonschema
from jsonschema.exceptions import ValidationError, SchemaError
import yaml

__version__ = "0.4.0"


class State(TypedDict):
    verbose: bool
    base_dir: Path


state: State = {
    "verbose": False,
    "base_dir": Path("."),
}


class Project(TypedDict):
    file: Path
    metadata: dict[str, Any]  # pyright: ignore[reportExplicitAny]
    content: str


projects: dict[str, Project] = {}

app = typer.Typer(pretty_exceptions_enable=False)
err_console = Console(stderr=True)

error_code = 1
errors: list[str] = []


def log_error(msg: str):
    errors.append(msg)


MD_FILE_DEFAULT = "index.md"


def init():
    if state["verbose"]:
        print(f"Initializing projects from {state['base_dir']}")
    md_files: list[Path] = []
    for file in state["base_dir"].glob("*"):
        if file.is_dir():
            full_path = file.resolve() / MD_FILE_DEFAULT
            if not full_path.exists() or not full_path.is_file():
                log_error(f"Missing {MD_FILE_DEFAULT} in {file}")
                continue
            md_files.append(full_path)

    if len(errors) != 0:
        return False

    md_files = sorted(md_files)

    projects.clear()
    for file in md_files:
        if state["verbose"]:
            print(f"Reading file {file}")
        with open(file, "r") as f:
            metadata, content = frontmatter.parse(f.read())
            title = cast(str, metadata["title"])
            if title in projects:
                log_error(
                    f"Duplicate project title: {title} in {projects[title]['file']} and {file}"
                )
                continue
            projects[title] = {
                "file": file,
                "metadata": metadata,
                "content": content,
            }

    if len(errors) != 0:
        return False

    return True


def parse_filter(filter: str | None) -> tuple[CodeType, set[str]] | tuple[None, None]:
    if filter is None:
        return None, None

    try:
        parsed = ast.parse(filter, mode="eval")
    except SyntaxError as e:
        log_error(f"Invalid filter: {e}")
        raise ValueError("Invalid filter")

    if (
        not isinstance(parsed.body, ast.BoolOp)
        and not isinstance(parsed.body, ast.UnaryOp)
        and not isinstance(parsed.body, ast.Name)
    ):
        if not isinstance(parsed.body, ast.Compare):
            log_error(f"Invalid filter: {filter}")
            raise ValueError("Invalid filter")

        final = ast.BoolOp(op=ast.And(), values=[parsed.body, ast.Constant(value=True)])
        parsed.body = ast.copy_location(final, parsed.body)
        parsed = ast.fix_missing_locations(parsed)

    compiled = compile(parsed, filename="<ast>", mode="eval")

    variables: set[str] = set()
    for node in ast.walk(parsed):
        if isinstance(node, ast.Name):
            variables.add(node.id)

    return compiled, variables


@app.command("list")
def list_projects(
    filter: Annotated[
        str | None, typer.Option("--filter", "-f", help="Filter by element")
    ] = None,
):
    try:
        compiled, variables = parse_filter(filter)
    except Exception:
        return False

    for p in projects.values():
        if compiled is not None and variables is not None:
            context: dict[str, Any] = {}  # pyright: ignore[reportExplicitAny]
            for v in variables:
                if v == "datetime":  # import datetime if used
                    context[v] = datetime
                    continue
                if v not in p["metadata"]:
                    context[v] = ""
                else:
                    context[v] = p["metadata"][v]
            result = cast(bool, eval(compiled, None, context))
            if not result:
                continue

        print(f"- {p['metadata']['title']}")


@app.command()
def show(name: str):
    selected: list[str] = []
    for p in projects.values():
        title = cast(str, p["metadata"]["title"])
        if name.lower() in title.lower():
            selected.append(title)

    if len(selected) == 0:
        log_error(f"Project {name} not found.")

    found = None
    for s in selected:
        if name.lower() == s.lower():
            found = s

    if len(selected) > 1:
        if found is None:
            print(f"Many projects found: {', '.join([p for p in selected])}")
            return False
    elif len(selected) == 1:
        found = selected[0]
    else:
        return False

    if state["verbose"]:
        print(f"Project found: {projects[found]['file']}")
    print(projects[found]["metadata"])


@app.command()
def validate(
    schema: Path,
    stop_on_error: Annotated[
        bool, typer.Option("--stop-on-error", "-s", help="Stop on first error")
    ] = False,
):
    with open(schema, "r") as f:
        sc = yaml.safe_load(f.read())  # pyright: ignore[reportAny]

    invalid: dict[str, str] = {}
    for p in projects.values():
        try:
            jsonschema.validate(p["metadata"], sc)  # pyright: ignore[reportAny]
        except ValidationError as e:
            title = cast(str, p["metadata"]["title"])
            invalid[title] = e.message
        except SchemaError as e:
            log_error(f"Invalid schema: {e}")
            return False

    if len(invalid) != 0:
        if stop_on_error:
            title, error = next(iter(invalid.items()))
            print(f"error for project {title}")
            print(error)
        else:
            for title, error in invalid.items():
                print(f"- {title}: {error}")
        return False

    if len(errors) != 0:
        return False

    print("All projects are valid.")


def version_callback(value: bool):
    if value:
        print(f"projects-organizer {__version__}")
        raise typer.Exit()


def result_cb(
    executed_command_result: Any,  # pyright: ignore[reportAny, reportExplicitAny]
    verbose: bool,  # pyright: ignore[reportUnusedParameter]
    base_dir: Path,  # pyright: ignore[reportUnusedParameter]
    version: bool,  # pyright: ignore[reportUnusedParameter]
):
    projects.clear()
    if len(errors) != 0:
        for e in errors:
            err_console.print(f"[red]error[/red]: {e}")
        errors.clear()
        raise typer.Exit(code=error_code)

    if executed_command_result is not None and executed_command_result is not True:
        raise typer.Exit(code=1)


@app.callback(result_callback=result_cb)
def main_options(
    verbose: Annotated[bool, typer.Option("--verbose", "-v")] = state["verbose"],
    base_dir: Annotated[Path, typer.Option("--base_dir", "-d")] = state["base_dir"],
    version: Annotated[  # pyright: ignore[reportUnusedParameter]
        bool | None, typer.Option("--version", callback=version_callback)
    ] = None,
):
    state["verbose"] = verbose
    state["base_dir"] = base_dir

    if not state["base_dir"].exists() and not state["base_dir"].is_dir():
        log_error(
            f"Base directory {state['base_dir']} does not exist or is not a directory."
        )
        return False

    return init()


if __name__ == "__main__":
    app()  # pragma: no cover



import pytest
import yaml


projects = {
    "project1": {
        "title": "Project 1",
        "created_at": "2023-01-01",
        "archived": True,
        "tags": ["dev", "python"],
    },
    "project2": {
        "title": "Project 2",
        "created_at": "2024-01-01",
        "tags": ["dev", "C++"],
    },
    "project3": {
        "title": "Project 3",
        "created_at": "2025-01-01",
        "tags": ["dev", "python"],
    },
}


schema = {
    "type": "object",
    "properties": {
        "title": {
            "type": "string",
        },
        "created_at": {
            "type": "string",
            "format": "date",
        },
        "archived": {
            "type": "boolean",
        },
        "tags": {
            "type": "array",
            "items": {"type": "string", "enum": ["dev", "python", "C++"]},
        },
    },
    "required": ["title"],
    "additionalProperties": False,
}


schema_missing_tag = {
    "type": "object",
    "properties": {
        "title": {
            "type": "string",
        },
        "created_at": {
            "type": "string",
            "format": "date",
        },
        "archived": {
            "type": "boolean",
        },
        "tags": {
            "type": "array",
            "items": {"type": "string", "enum": ["dev", "C++"]},
        },
    },
    "required": ["title"],
    "additionalProperties": False,
}


schema_invalid = {
    "type": "object",
    "properties": {
        "title": "foo",
        "date": {
            "type": "string",
            "format": "date",
        },
        "archived": {
            "type": "boolean",
        },
        "tags": {
            "type": "array",
            "items": {"type": "string", "enum": ["dev", "C++"]},
        },
    },
    "required": ["title"],
    "additionalProperties": False,
}


def pytest_configure():
    pytest.projects = projects


def create_projects_to_dir(projects, p_dir):
    for name, data in projects.items():
        p_subdir = p_dir / name
        if not p_subdir.exists():
            p_subdir.mkdir()
        with open(p_subdir / "index.md", "w") as f:
            f.write("---\n")
            f.write(
                yaml.dump(data, indent=2, default_flow_style=False, sort_keys=False)
            )
            f.write("---\n")
            f.write("My description")
    return p_dir


@pytest.fixture(scope="session")
def projects_dir(tmp_path_factory):
    return create_projects_to_dir(projects, tmp_path_factory.mktemp("projects_valid"))


@pytest.fixture(scope="session")
def projects_dir_empty_project(tmp_path_factory):
    dir = create_projects_to_dir(
        projects, tmp_path_factory.mktemp("projects_with_empty")
    )
    p_subdir = dir / "project_empty"
    if not p_subdir.exists():
        p_subdir.mkdir()
    return dir


@pytest.fixture(scope="session")
def projects_dir_duplicate_project(tmp_path_factory):
    import copy

    projects_dup = copy.deepcopy(projects)
    projects_dup["project_dup"] = projects_dup["project1"]
    return create_projects_to_dir(
        projects_dup, tmp_path_factory.mktemp("projects_with_dup")
    )


def write_data_to_file(data, filename):
    with open(filename, "w") as f:
        f.write(yaml.dump(data, indent=2, default_flow_style=False, sort_keys=False))
    return filename


@pytest.fixture(scope="session")
def schema_file(projects_dir):
    return write_data_to_file(schema, projects_dir / "schema.yaml")


@pytest.fixture(scope="session")
def schema_file_missing_tag(projects_dir):
    return write_data_to_file(schema_missing_tag, projects_dir / "invalid_schema.yaml")


@pytest.fixture(scope="session")
def schema_file_invalid(projects_dir):
    return write_data_to_file(schema_invalid, projects_dir / "invalid_schema.yaml")



from projects_organizer import app, __version__
from typer.testing import CliRunner


def test_version():
    runner = CliRunner()
    test_args = ["--version"]
    result = runner.invoke(app, test_args)
    assert result.exit_code == 0
    assert result.stdout == f"projects-organizer {__version__}\n"


def test_main_invalid_dir():
    runner = CliRunner(mix_stderr=False)
    test_args = ["-d", "not_exists", "list"]
    result = runner.invoke(app, test_args)
    assert result.exit_code == 1
    assert result.stdout == ""
